------------------------------------------------------------------------------
--- A HtmlServer manages the event handlers used in
--- dynamic web pages on the server side.
--- Each event handler is encoded as a string and stored on the server side
--- with a unique key. Only this key is sent in the actual web page
--- to the client.
--- Event handlers are only valid for a particular time period
--- specified by <code>eventHandlerExpiration</code>, i.e., after that time
--- event handlers will be deleted.

module HtmlServer(storeHandlerStringsWithCgiKey,retrieveHandlerStringWithCgiKey) where

import Ports
import Read(readNat)
import ReadShowTerm(readsQTerm)
import Time
import System

-- Stores a list of new event handlers for a given cgi program and the corresponding
-- arguments with a new key.
storeHandlerStringsWithCgiKey :: String -> [(String,String)] -> [(String,String)] -> IO ()
storeHandlerStringsWithCgiKey cgikey oldcenv handlerkeys = do
  time <- getClockTime
  let (msgs,unifs) = generateEventServerMessages (toDate (eventHandlerExpiration time))
                                                 handlerkeys
  sendEventServerMessages (CleanOldEventHandlers:msgs)
  doSolve (foldr (&) success unifs)
 where
   generateEventServerMessages _ [] = ([],[])
   generateEventServerMessages expiredate ((handlerstring,newkey):evhs) =
     let (msgs,unifs) = generateEventServerMessages expiredate evhs
      in (StoreNewEnvEventWithCgiKey key expiredate cgikey oldcenv handlerstring : msgs,
          (show key ++ ' ':show expiredate =:= newkey) : unifs)
    where key free

-- Retrieves a previously stored event handler for a cgi program.
-- Returns Nothing if the handler is no longer available, i.e., expired.
retrieveHandlerStringWithCgiKey :: String -> String
                                   -> IO (Maybe ([(String,String)],String))
retrieveHandlerStringWithCgiKey key cgikey = 
  let (nums,dates) = break (==' ') key
   in if null dates
      then return Nothing
      else let datel = readsQTerm dates
            in if null datel
               then return Nothing
               else sendEventServerMessages
                     [GetEnvEventWithCgiKey (readNat nums) (fst (head datel)) cgikey info] >>
                    return info
 where info free

-- Define for a given date a new date when the event handler expires.
eventHandlerExpiration :: ClockTime -> ClockTime
eventHandlerExpiration = addHours 1
--eventHandlerExpiration = addMinutes 1

-- Define for a given date a new date when the next cleanup of event handlers
-- should be done.
nextCleanup :: ClockTime -> ClockTime
nextCleanup = addMinutes 5

---------------------------------------------------------------------------
-- The implementation of the event handler server via ports:

type Date = (Int,Int,Int,Int,Int,Int)

-- Messages for event handler server:
data EHSMsg = StoreNewEventWithCgiKey Int Date String String
            | GetEventWithCgiKey Int Date String (Maybe String)
            | StoreNewEnvEventWithCgiKey Int Date String [(String,String)] String
            | GetEnvEventWithCgiKey Int Date String (Maybe ([(String,String)],String))
            | CleanOldEventHandlers
            | ShowAllHandlers String
            | SketchAllHandlers Int String
            | Stop

-- Start the event handler server via script:
startServer = do
  pakcshome <- getEnviron "PAKCSHOME"
  system (pakcshome++"/www/start")

-- Stop the event handler server (usually, not required):
stopServer = sendEventServerMessages [Stop]

-- show all currently stored handlers:
showAllHandlers = do
  sendEventServerMessages [ShowAllHandlers s]
  putStr s
 where s free

-- sketch all currently stored handlers up to a particular length:
sketchAllHandlers len = do
  sendEventServerMessages [SketchAllHandlers len s]
  putStr s
 where s free


-- Open the external port "ehserver" and start the event handler server:
serve = do
  msgs <- openNamedPort "ehserver"
  time <- getClockTime
  ehServer 0 (toDate time) [] msgs

-- The event handler server loop:
-- Argument 1: Current index for numbering new events
-- Argument 2: Next date when cleanup is necessary
-- Argument 3: The current event handlers
--             (index,experiation date,cgikey,oldcenv,contents)
ehServer :: Int -> Date -> [(Int,Date,String,([(String,String)],String))] -> [EHSMsg]
            -> IO ()

ehServer _ _ _ (Stop:_) = done

ehServer maxkey cleandate ehs
         (StoreNewEnvEventWithCgiKey key time cgikey oldcenv info : msgs) = do
  doSolve $ key =:= maxkey
  ehServer incMaxKey cleandate ((key,time,cgikey,(oldcenv,info)):ehs) msgs
 where
  incMaxKey = if maxkey>30000 then 0 else maxkey+1 -- to avoid overflows

ehServer maxkey cleandate ehs (GetEnvEventWithCgiKey key time cgikey mbinfo : msgs)
  | mbinfo =:= searchEv ehs
  = ehServer maxkey cleandate ehs msgs
 where
   searchEv [] = Nothing
   searchEv ((n,t,c,i):es) = if key==n && time==t
                             then if c==cgikey then Just i else Nothing
                             else searchEv es

ehServer maxkey cleandate ehs -- to be deleted in the future
         (StoreNewEventWithCgiKey key time cgikey info : msgs) = do
  doSolve $ key =:= maxkey
  ehServer incMaxKey cleandate ((key,time,cgikey,([],info)):ehs) msgs
 where
  incMaxKey = if maxkey>30000 then 0 else maxkey+1 -- to avoid overflows

ehServer maxkey cleandate ehs  -- to be deleted in the future
         (GetEventWithCgiKey key time cgikey mbinfo : msgs)
  | mbinfo =:= searchEv ehs
  = ehServer maxkey cleandate ehs msgs
 where
   searchEv [] = Nothing
   searchEv ((n,t,c,(_,i)):es) = if key==n && time==t
                                 then if c==cgikey then Just i else Nothing
                                 else searchEv es

ehServer maxkey cleandate ehs (CleanOldEventHandlers : msgs) = do
  time <- getClockTime
  if compareDate (toDate time) cleandate == LT
   then ehServer maxkey cleandate ehs msgs
   else putStrLn (toDateString time ++": cleanup (current number of handlers: "++
                  show (length ehs) ++")") >>
        ehServer maxkey (toDate (nextCleanup time))
                 (filter (isNotExpired (toDate time)) ehs) msgs
 where
  isNotExpired time (_,etime,_,_) = compareDate time etime == LT

-- bind argument to a string representation of all event handlers
ehServer maxkey cleandate ehs (ShowAllHandlers s : msgs)
  | s =:= "Next cleanup: " ++ toDateString (dateToClockTime cleandate) ++"\n"++
          "Current event handlers:\n" ++ concatMap showWithoutCgiKey ehs
  = ehServer maxkey cleandate ehs msgs
 where
  showWithoutCgiKey (key,time,_,(_,info)) =
    "No. "++ show key ++" "++ show time ++": "++info++"\n"
  showWithCgiKey (key,time,cgi,(oldcenv,info)) = -- only for testing
    "No. "++ show key ++" / date: "++ show time ++" / key: \"" ++cgi++"\":\n  "++
    show oldcenv ++ "\n  "++info++"\n"

-- bind argument to a short string representation of all event handlers
-- where from each event handler only the first len characters are shown:
ehServer maxkey cleandate ehs (SketchAllHandlers len s : msgs)
  | s =:= "Next cleanup: " ++ toDateString (dateToClockTime cleandate) ++"\n"++
          "Current event handlers:\n" ++ concatMap showWithoutCgiKey ehs
  = ehServer maxkey cleandate ehs msgs
 where
  showWithoutCgiKey (key,time,_,(_,info)) =
    let infolength = length info in
    "No. "++ show key ++" "++ show time ++": "++
    take len info ++
    (if len<infolength then "...("++show infolength++" bytes)" else "") ++"\n"

-- The timeout (in msec) of the event handler server.
-- If the event handler server does not answer within the timeout period,
-- we assume that the server does not exist and we start a new one:
eventHandlerServerTimeOut = 2000

-- send a list of messages to local event handler server:
sendEventServerMessages :: [EHSMsg] -> IO ()
sendEventServerMessages msgs =
  connectPortRepeat eventHandlerServerTimeOut done 1 "ehserver@localhost" >>=
  maybe (startServer >> sendEventServerMessages msgs)
        (\port->mapIO_ (\m->doSend m port) msgs)

